/*
 * Copyright (C) 2004-2008  See the AUTHORS file for details.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation.
 */

#include "znc.h"
#include "User.h"
#include <pwd.h>
#include <grp.h>

class CChroot : public CGlobalModule
{
public:
	GLOBALMODCONSTRUCTOR(CChroot) {}

	virtual ~CChroot() {}

	uid_t GetUser(const CString& sUser, CString& sMessage) {
		uid_t ret = sUser.ToUInt();
		if (ret != 0)
			return ret;

		struct passwd *pUser = getpwnam(sUser.c_str());
		if (!pUser) {
			sMessage = "User [" + sUser + "] not found!";
			return 0;
		}

		return pUser->pw_uid;
	}

	gid_t GetGroup(const CString& sGroup, CString& sMessage) {
		gid_t ret = sGroup.ToUInt();
		if (ret != 0)
			return ret;

		struct group *pGroup = getgrnam(sGroup.c_str());
		if (!pGroup) {
			sMessage = "Group [" + sGroup + "] not found!";
			return 0;
		}
		return pGroup->gr_gid;
	}

	void fixDNS() {
		// The first time some DNS routines are used, some resolver lib
		// is loaded. After we chrooted, that lib is no longer found,
		// so load it now.
		struct addrinfo *res = NULL;

		getaddrinfo( "znc.in", NULL, NULL, &res );
		if (res)
			freeaddrinfo( res );
	}

	CString GetChroot() {
		return m_sChroot;
	}

	virtual bool OnLoad(const CString& sArgs, CString& sMessage) {
		CString sUser = sArgs.Token(0);
		CString sGroup = sArgs.Token(1, true);
		m_sChroot = CZNC::Get().GetZNCPath();

		if (sGroup.empty()) {
			sMessage = "Usage: LoadModule = Chroot <user> <group>";
			return false;
		}

		m_user = GetUser(sUser, sMessage);
		if (m_user == 0)
			return false;

		m_group = GetGroup(sGroup, sMessage);
		if (m_group == 0)
			return false;

		fixDNS();

		sMessage.clear();

#if HAVE_LIBSSL
		if (!CFile::Exists(GetChroot() + "/dev/urandom")) {
			sMessage = "If you plan to use SSL, it is recommended for"
				" security reasons to create [" + GetChroot() +
				"/dev/urandom] via [mknod -m 644 urandom c 1 9]";
		}
#endif

		return true;
	}

	void FixModule(CModule* pMod) {
		CString sChroot = GetChroot();
		CString sDataDir = pMod->GetModDataDir();
		CString sSavePath = pMod->GetSavePath();

		if (!sDataDir.TrimPrefix(sChroot)) {
			CUtils::PrintError("The datadir for [" + pMod->GetModName()
					+ "] is outside of the chroot. This might "
					"render the module unusable.");
		}
		if (!sSavePath.TrimPrefix(sChroot)) {
			// This one actually can't happen, but well...
			CUtils::PrintError("The save path for [" + pMod->GetModName()
					+ "] is outside of the chroot. This might "
					"render the module unusable.");
		}

#warning This module needs changes in ZNC core which are not and will not
#warning be included. Better read the chroot page in the wiki than use this.
#if 0
		pMod->SetModDataDir(sDataDir);
		pMod->SetSavePath(sSavePath);
#endif
	}

	void FixGlobalModules() {
		CGlobalModules& vMods = CZNC::Get().GetModules();
		CGlobalModules::iterator it;
		for (it = vMods.begin(); it != vMods.end(); it++) {
			FixModule(*it);
		}
	}

	void FixUserModules() {
		const map<CString, CUser*>& vUsers = CZNC::Get().GetUserMap();
		map<CString, CUser*>::const_iterator user_it;

		for (user_it = vUsers.begin(); user_it != vUsers.end(); user_it++) {
			CModules& vMods = user_it->second->GetModules();
			CModules::iterator it;

			for (it = vMods.begin(); it != vMods.end(); it++) {
				FixModule(*it);
			}
		}
	}

	virtual bool OnBoot() {
		int i;

		CUtils::PrintAction("Chrooting and dropping permissions");

		// Chroot into our datadir
		i = chroot(CZNC::Get().GetZNCPath().c_str());
		if (i < 0) {
			CUtils::PrintStatus(false, "Could not change to new root dir! ["
					+ CString(strerror(errno)) + "]");
			return false;
		}

		i = chdir("/");
		if (i < 0) {
			CUtils::PrintStatus(false, "Could not cd to new root dir! ["
					+ CString(strerror(errno)) + "]");
			return false;
		}

		// Clear all the supplementary groups
		i = setgroups(0, NULL);
		if (i < 0) {
			CUtils::PrintStatus(false, "Could not remove supplementary groups! ["
					+ CString(strerror(errno)) + "]");
			return false;
		}

		// Set the group (if we are root, this sets all three group IDs)
		i = setgid(m_group);
		if (i < 0) {
			CUtils::PrintStatus(false, "Could not switch group id! ["
					+ CString(strerror(errno)) + "]");
			return false;
		}

		// and set the user (if we are root, this sets all three user IDs)
		i = setuid(m_user);
		if (i < 0) {
			CUtils::PrintStatus(false, "Could not switch user id! ["
					+ CString(strerror(errno)) + "]");
			return false;
		}

		CUtils::PrintStatus(true);

		// Fix our internal paths
		putenv("HOME=/");
		CZNC::Get().InitDirs("", "/");

		FixGlobalModules();
		FixUserModules();

		return true;
	}

	virtual void OnPreRehash() {
		CZNC::Get().Broadcast("Rehashing might not work in a chroot!", true);
	}

protected:
	CString	m_sChroot;
	uid_t	m_user;
	gid_t	m_group;
};

GLOBALMODULEDEFS(CChroot, "Chroot ZNC to its data dir and switch to a different user")
